# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2019. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use

import requests
import logging
from resilient import is_env_proxies_set, get_and_parse_proxy_env_var, constants as res_constants
from resilient_lib.components.integration_errors import IntegrationError
from resilient_lib.components.resilient_common import str_to_bool
from deprecated import deprecated


LOG = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 30


class RequestsCommon(object):
    """
    This class represents common functions around the use of the requests package for REST based APIs.
    It incorporates the app.config section ``[integrations]`` which can be used to define a common set of proxies
    for use by all functions using this library.

    Any similar properties in the function’s section would override the [integrations] properties.

    .. note::
      In the Atomic Function template, as of version 41.1, :class:`RequestsCommon` is instantiated and available
      in a class that inherits :class:`~resilient_circuits.app_function_component.AppFunctionComponent` as an ``rc`` attribute:

      .. code-block:: python

         response = self.rc.execute(method="GET", url="ibm.com")

    :param opts: all configurations found in the app.config file
    :type opts: dict
    :param function_opts: all configurations found in the ``[my_function]`` section of the app.config file
    :type function_opts: dict
    """
    def __init__(self, opts=None, function_opts=None):
        # capture the properties for the integration as well as the global settings for all integrations for proxy urls
        self.integration_options = opts.get("integrations", None) if opts else None
        self.function_opts = function_opts

        # base class requests object (i.e. with persistent sessions):
        self.request_obj = requests.Session()

    def get_proxies(self):
        """
        Proxies can be specified globally for all integrations or specifically per function.

        * If the environmental variables HTTPS_PROXY, HTTP_PROXY and NO_PROXY
          are set, this returns ``None``, as if for a `requests.Request <https://docs.python-requests.org/en/latest/api/#requests.Request>`_.
          If ``proxies`` are ``None``, the environmental variables are used
        * If ``http_proxy`` or ``https_proxy`` is set in the **Function Section** (``[my_function]``) of your app.config file,
          returns a dictionary mapping protocol to the URL of the proxy.
        * If ``http_proxy`` or ``https_proxy`` is set in the **Integrations Section** (``[integrations]``) of your app.config file,
          returns a dictionary mapping protocol to the URL of the proxy.

        :return: A dictionary mapping protocol to the URL of the proxy or ``None``
        :rtype: dict or ``None``
        """
        proxies = None

        if is_env_proxies_set():
            proxy_details = get_and_parse_proxy_env_var(res_constants.ENV_HTTPS_PROXY)

            if not proxy_details:
                proxy_details = get_and_parse_proxy_env_var(res_constants.ENV_HTTP_PROXY)

            if proxy_details:
                LOG.debug(u"Sending request through proxy: '%s://%s:%s'", proxy_details.get("scheme"), proxy_details.get("hostname"), proxy_details.get("port"))

            return proxies

        if self.integration_options and (self.integration_options.get("http_proxy") or self.integration_options.get("https_proxy")):
            proxies = {'http': self.integration_options.get("http_proxy"), 'https': self.integration_options.get("https_proxy")}

        if self.function_opts and (self.function_opts.get("http_proxy") or self.function_opts.get("https_proxy")):
            proxies = {'http': self.function_opts.get("http_proxy"), 'https': self.function_opts.get("https_proxy")}

        return proxies

    def get_timeout(self):
        """
        Get the default timeout value in the **Function Section** (``[my_function]``) of your app.config file,
        or from the **Integrations Section** (``[integrations]``) of your app.config file.

        ``[my_function]`` section takes precedence over ``[integrations]`` section value.

        :return: timeout value
        :rtype: int
        """
        timeout = DEFAULT_TIMEOUT

        if self.integration_options and self.integration_options.get("timeout"):
            timeout = self.integration_options.get("timeout")

        if self.function_opts and self.function_opts.get("timeout"):
            timeout = self.function_opts.get("timeout")

        return int(timeout)

    def get_client_auth(self):
        """
        A client certificate for authenticating calls to external endpoints can be specified on a per function basis.

        If ``client_auth_cert`` and ``client_auth_key`` are set in the **Function Section** (``[my_function]``) of your app.config file,
        returns a tuple containing the respective paths to the certificate and private key for the client cert.

        Example:

          .. code-block::

            [my_function]
            ...
            client_auth_cert = <path_to_cert.pem>
            client_auth_key = <path_to_cert_private_key.pem>

        :return: The filepaths for the client side certificate and the private key as a tuple of both files' paths
                 or ``None`` if either one of the values are missing
        :rtype: tuple(str, str)
        """
        cert = None

        if self.function_opts and (self.function_opts.get("client_auth_cert") and self.function_opts.get("client_auth_key")):
            cert = (self.function_opts.get("client_auth_cert"), self.function_opts.get("client_auth_key"))

        return cert

    def get_verify(self):
        """
        Get ``verify`` parameter from app config or from env var
        ``REQUESTS_CA_BUNDLE`` which is the default way to set
        verify with python requests library.

        Value can be set in ``[integrations]`` or in the ``[fn_my_app]`` section

        Value in ``[fn_my_app]`` takes precedence over ``[integrations]``
        which takes precedence over ``REQUESTS_CA_BUNDLE``

        :param app_options: App config dict
        :type app_options: dict
        :return: Value to set ``requests.session.verify`` to.
            Either a path or a boolean
        :rtype: bool or str or None (which will default to requests default which is ``True``)
        """
        verify = None

        # look in app's config section first
        if self.function_opts:
            verify = self.function_opts.get("verify", True)

        # if not found in config, then [integration] wide section can be used;
        # NOTE: this is pretty much limited to just integration server envs
        # as most app host deployments will just be the single app, and thus
        # an integration's wide setting section is not relevant
        if verify is None and self.integration_options:
            verify = self.integration_options.get("verify", True)

        # convert potential strings to boolean if necessary
        # NOTE: it is possible that a string path should be returned,
        # in which case we don't want there to be a boolean conversion
        if isinstance(verify, str) and verify.lower() in ["false", "true"]:
            verify = str_to_bool(verify)

        return verify

    # alias for existing instances of ``get_clientauth``, but from now on
    # should use ``get_client_auth``
    get_clientauth = get_client_auth


    def execute(self, method, url, timeout=None, proxies=None, callback=None, clientauth=None, verify=None, **kwargs):
        """
        Constructs and sends a request. Returns a
        `requests.Response <https://docs.python-requests.org/en/latest/api/#requests.Response>`_ object.

        This uses the ``requests.request()`` function to
        make a call. The inputs are mapped to this function.
        See `requests.request() <https://docs.python-requests.org/en/latest/api/#requests.request>`_
        for information on any parameters available, but not documented here.

        :param method: Rest method to execute (``GET``, ``POST``, etc...)
        :type method: str
        :param url: URL to execute request against
        :type url: str
        :param timeout: (Optional) Number of seconds to wait for the server to send data
            before sending a float or a timeout tuple (connect timeout, read timeout).
            *See requests docs for more*. If ``None`` it looks in the ``[integrations]``
            section of your app.config for the ``timeout`` setting.
        :type timeout: float or tuple
        :param proxies: (Optional) Dictionary mapping protocol to the URL of the proxy.
            The mapping protocol must be in the format:

            .. code-block::

                {
                    "https_proxy": "https://localhost:8080,
                    "http_proxy": "http://localhost:8080
                }
        :type proxies: dict
        :param callback: (Optional) Once a response is received from the endpoint,
            return this callback function passing in the ``response`` as its
            only parameter. Can be used to specifically handle errors.
        :type callback: function
        :param clientauth: (Optional) Equivalent to the ``cert`` parameter of ``requests``.
            Client-side certificates can be configured automatically in
            the app's section of the app.config. These values will be read in automatically by
            ``RequestsCommon`` when ``execute`` is called.

            Example:

            .. code-block::

                [resilient]
                ... # configs for connection to SOAR

                [fn_my_app]
                ... # some other configs
                client_auth_cert=<path_to_client_auth_cert_file>
                client_auth_key=<path_to_client_auth_private_key_file>

            If not set in the app.config file, the filepath for the client side certificate and the
            private key can be passed to this function either as the path to a single file or as a
            tuple of both files' paths.
        :type clientauth: str or tuple(str, str)
        :param verify: (Optional) Either a boolean, in which case it controls whether requests verifies
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``None`` as passed in here. If ``None``,
            then the value is searched for in the app's section of the app.config. If no value
            is found, then the requests.request() default of ``True`` is used. In that case,
            the CA bundle found at ``requests.utils.DEFAULT_CA_BUNDLE_PATH`` is the CA bundle used,
            which is usually a bundle provided by Mozilla. Setting ``verify=True`` is the safest option
            and should only be changed if you have a self-signed certificate or you want to bypass
            SSL for testing purposes. A production level app should never disable SSL verification
            completely.

            Example:

            .. code-block::

                [fn_my_app]
                ... # some other configs
                verify=<path/to/CA/bundle/to/use>

            NOTE: The value held in the app's config will be overwritten if a value is passed in by the call
            to ``execute``
        :type verify: bool or str
        :return: the ``response`` from the endpoint or return ``callback`` if defined.
        :rtype: `requests.Response <https://docs.python-requests.org/en/latest/api/#requests.Response>`_ object
            or ``callback`` function.
        """
        try:
            if method.lower() not in ('get', 'post', 'put', 'patch', 'delete', 'head', 'options'):
                raise IntegrationError("unknown method {}".format(method))

            # If proxies was not set check if they are set in the config
            if not proxies:
                proxies = self.get_proxies()

            if not timeout:
                timeout = self.get_timeout()

            if not clientauth:
                clientauth = self.get_client_auth()

            if verify is None:
                verify = self.get_verify()

            # Log the parameter inputs that are not None
            args_dict = locals()
            # When debugging execute_call_v2 in PyCharm you may get an exception when executing the for-loop:
            # Dictionary changed size during iteration.
            # To work around this while debugging you can change the following line to:
            # args = list(args_dict.keys())
            args = args_dict.keys()
            for k in args:
                if k != "self" and k != "kwargs" and args_dict[k] is not None:
                    LOG.debug("  %s: %s", k, args_dict[k])

            # Pass request to requests.request() function
            response = self.request_obj.request(method, url, timeout=timeout, proxies=proxies, cert=clientauth, verify=verify, **kwargs)

            # Debug logging
            LOG.debug(response.status_code)
            LOG.debug(response.content)

            # custom handler for response handling
            # set callback to be the name of the method you would like to call
            # to do your custom error handling and return the response
            if callback:
                return callback(response)

            # Raise error is bad status code is returned
            response.raise_for_status()

            # Return requests.Response object
            return response

        except Exception as err:
            msg = str(err)
            LOG.error(msg)
            raise IntegrationError(msg)

    # Create alias for execute_call_v2
    execute_call_v2 = execute

    @deprecated(version="v34.0", reason="Use the new method execute()")
    def execute_call(self, verb, url, payload={}, log=None, basicauth=None, verify_flag=True, headers=None,
                     proxies=None, timeout=None, resp_type='json', callback=None):
        """
        Function: perform the http API call. Different types of http operations are supported:
        GET, HEAD, PATCH, POST, PUT, DELETE
        Errors raise IntegrationError
        If a callback method is provided, then it's called to handle the error

        When using argument 'json' Content-Type will automatically be set to "application/json" by requests lib.

        :param verb: GET, HEAD, PATCH, POST, PUT, DELETE
        :param url:
        :param basicauth: used for basic authentication - (user, password)
        :param payload:
        :param log: optional log statement
        :param verify_flag: True/False - False used for debugging generally
        :param headers: dictionary of http headers
        :param proxies: http and https proxies for call
        :param timeout: timeout before call should abort
        :param resp_type: type of output to return: json, text, bytes
        :param callback: callback routine used to handle errors
        :return: json of returned data
        """

        try:
            (payload and log) and log.debug(payload)

            if verb.lower() not in ('get', 'post', 'put', 'patch', 'delete', 'head'):
                raise IntegrationError("unknown verb {}".format(verb))

            if proxies is None:
                proxies = self.get_proxies()

            if verb.lower() == 'post':

                content_type = get_case_insensitive_key_value(headers, "Content-Type")

                if is_payload_in_json(content_type):
                    resp = requests.request(verb.upper(), url, verify=verify_flag, headers=headers, json=payload,
                                            auth=basicauth, timeout=timeout, proxies=proxies)
                else:
                    resp = requests.request(verb.upper(), url, verify=verify_flag, headers=headers, data=payload,
                                            auth=basicauth, timeout=timeout, proxies=proxies)
            else:
                resp = requests.request(verb.upper(), url, verify=verify_flag, headers=headers, params=payload,
                                        auth=basicauth, timeout=timeout, proxies=proxies)

            if not resp:
                raise IntegrationError('no response returned')

            # custom handler for response handling?
            if callback:
                return callback(resp)

            # standard error handling
            if resp.status_code >= 300:
                # get the result
                # log resp.status_code in case resp.text isn't available
                raise IntegrationError(
                    "status_code: {}, msg: {}".format(resp.status_code, resp.text if resp.text else "N/A"))

            # check if anything returned
            log and log.debug(resp.text)

            # get the result
            if resp_type == 'json':
                r = resp.json()
            elif resp_type == 'text':
                r = resp.text
            elif resp_type == 'bytes':
                r = resp.content
            else:
                raise IntegrationError("incorrect response type: {}".format(resp_type))

            # Produce a IntegrationError with the return value
            return r      # json object needed, not a string representation
        except Exception as err:
            msg = str(err)
            log and log.error(msg)
            raise IntegrationError(msg)

class RequestsCommonWithoutSession(RequestsCommon):
    """
    This class extends :class:`RequestsCommon` maintaining the behavior of 
    ``RequestsCommon`` that was present in versions <= 47.1.x.

    The difference is that every time :class:`RequestsCommonWithoutSession.execute()`
    is called, this class will use ``requests.request`` to execute the request, while
    :class:`RequestsCommon.execute()` uses ``Session().request``.

    This class is interchangeable with :class:`RequestsCommon` and code that uses
    :class:`RequestsCommon` can be simiply refactored to use :class:`RequestsCommonWithoutSession`

    :param args: positional arguments matching :class:`RequestsCommon`
    :type args: list
    :param kwargs: named keyword arguments matching :class:`RequestsCommon`
    :type kwargs: dict
    """
    def __init__(self, *args, **kwargs):
        super(RequestsCommonWithoutSession, self).__init__(*args, **kwargs)
        self.request_obj = requests


def is_payload_in_json(content_type):
    """
    Verify the content_type.
    If "Content-Type" is NOT specified pass the payload to the "json" argument - return True.

    If "Content-Type" is specified:
        - if the value is "application/json" pass the payload to the "json" argument - return True.
        - if the value is NOT "application/json" pass the payload to the "data" argument - return False.
    :param content_type:
    :return: True or False
    """

    if not content_type:
        return True

    return "application/json" in content_type.lower()


def get_case_insensitive_key_value(dictionary, key):
    """
    Get case insensitive key value from dictionary.
    :param dictionary:
    :param key:
    :return: value or None
    """
    if dictionary is None:
        return None

    return next((v for k, v in dictionary.items() if k.lower() == key.lower()), None)



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2018. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use

import datetime
import io
import logging
import mimetypes
import os
import sys
import tempfile
if sys.version_info.major >= 3:
    from urllib.parse import quote
else:
    from urllib import quote

import resilient
from bs4 import BeautifulSoup
from cachetools import TTLCache, cached
from resilient_lib.util import constants
from six import string_types

CP4S_PREFIX = "cases-rest."
CP4S_RESOURCE_PREFIX = "app/respond"
INCIDENT_FRAGMENT = "#incidents"
CASE_FRAGMENT = "#cases"
TASK_FRAGMENT = "taskId="
TASK_DETAILS_FRAGMENT = "tabName=details"
PAYLOAD_VERSION = "1.0"

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)
LOG.addHandler(logging.StreamHandler())


def build_incident_url(url, incidentId, orgId=None):
    """
    Build the url to link to a SOAR incident or CP4S case.
    Add ``https`` if http/https is not provided at the start.
    If ``url`` is not a string, returns back the value given.

    ``orgId`` is optional to maintain backward compatibility, however, it is
    strongly recommended to provide the organization ID of the incident
    so that links work without unexpected hiccups when multiple orgs are
    available on your SOAR instance

    Returns a URL in the format ``https://<url>/#<incident_id>?orgId=<orgId>``.

    :param url: the URL of your SOAR instance
    :type url: str
    :param incidentId: the id of the incident
    :type incidentId: str|int
    :param orgId: (optional) the id of the org the incident lives in. If the user is logged into a different org
        and this is not set, the link produced may direct the user to a different incident resulting in
        unexpected results
    :type orgId: str|int|None
    :return: full URL to the incident
    :rtype: str
    """

    if not isinstance(url, string_types):
        LOG.warning("Called 'build_incident_url' with a '{0}'  but was expecting a 'str' URL value. Returning original value.".format(type(url)))
        return url

    # determine if host url needs http/s prefix
    # if not given, assumes https
    if not url.lower().startswith("http"):
        url = "https://{0}".format(url)

    # remove cp4s prefix if still present
    if CP4S_PREFIX in url:
        url = url.replace(CP4S_PREFIX, "")

        # unfortunately we can't insert app/respond unless we know the cp4s prefix was there
        # so we insert if missing
        # otherwise we have to assume this is a standalone instance link
        if CP4S_RESOURCE_PREFIX not in url:
            url = '/'.join([url, CP4S_RESOURCE_PREFIX])

    if CP4S_RESOURCE_PREFIX in url:
        fragment = CASE_FRAGMENT
    else:
        fragment = INCIDENT_FRAGMENT

    link = '/'.join([url, fragment, str(incidentId)])
    
    if orgId and isinstance(orgId, (str, int)):
        orgId = quote(str(orgId))
        link += "?orgId={0}".format(orgId)
    else:
        LOG.warning(constants.WARN_BUILD_INCIDENT_ORG_ID)

    return link


def build_task_url(url, incident_id, task_id, org_id):
    """
    Build the url to link to a SOAR/CP4S task.
    Add ``https`` if http/https is not provided at the start.
    If ``url`` is not a string, returns back the value given.

    Returns a URL in the format ``https://<url>/#<incident_id>?orgId=<org_id>&taskId=<task_id>&tabName=details``.

    :param url: the URL of your SOAR instance
    :type url: str
    :param incident_id: the id of the incident
    :type incident_id: str|int
    :param task_id: the id of the task
    :type task_id: str|int
    :param org_id: the id of the org the incident lives in
    :type org_id: str|int
    :return: full URL to the task's details tab
    :rtype: str
    """

    if not isinstance(url, string_types):
        LOG.warning("Called 'build_task_url' with a '{0}'  but was expecting a 'str' URL value. Returning original value.".format(type(url)))
        return url

    url = "{0}&{1}{2}&{3}".format(build_incident_url(url, incident_id, orgId=org_id), TASK_FRAGMENT, str(task_id), TASK_DETAILS_FRAGMENT)

    return url


def build_resilient_url(host, port):
    """
    Build basic url to SOAR/CP4S instance.
    Add 'https' if http/https is not provided at the start.
    If ``host`` is not a string, returns back the value given.

    :param host: host name
    :type host: str
    :param port: port
    :type port: str|int
    :return: base url
    :rtype: str
    """

    if not isinstance(host, string_types):
        LOG.warning("Called 'build_resilient_url' with a '{0}'  but was expecting a 'str' host value. Returning original value.".format(type(host)))
        return host

    # determine if host url needs http/s prefix
    # if not given, assumes https
    if not host.lower().startswith("http"):
        host = "https://{0}".format(host)

    # check if host is CP4S host
    if CP4S_PREFIX in host:
        host = host.replace(CP4S_PREFIX, "")
        return "{0}:{1}/{2}".format(host, port, CP4S_RESOURCE_PREFIX)

    return "{0}:{1}".format(host, port)


def clean_html(html_fragment):
    """
    SOAR textarea fields return HTML fragments. This routine removes the
    HTML and inserts any code within ``<div></div>`` with a linefeed.

    .. note::
        The string returned from this method may not format well as no presentation of line feeds are preserved,
        tags such as ``<br>`` or ``<ol>``, ``<ul>``, etc. are removed. See :class:`.MarkdownParser` class for a better way to translate
        html input to markdown.

    :param html_fragment: the html to clean up
    :type html_fragment: str
    :return: cleaned up code
    :rtype: str
    """

    if not html_fragment or not isinstance(html_fragment, string_types):
        return html_fragment

    s = BeautifulSoup(unescape(html_fragment), "html.parser")

    return ' '.join(s.strings)


def unescape(data):
    """
    Return unescaped data such as:
        * ``&gt;`` converts to ``>``
        * ``&quot`` converts to ``'``

    :param data: text to convert
    :type data: str
    :return: the text unescaped
    :rtype: str
    """
    if data is None:
        return None

    if sys.version_info.major < 3:
        # In PY 2, unescape is part of HTMLParser
        from HTMLParser import HTMLParser
        h = HTMLParser()
        return h.unescape(data)

    else:
        # In PY 3, unescape is in html library
        import html
        return html.unescape(data)


def validate_fields(field_list, kwargs):
    """
    Ensure each mandatory field in ``field_list`` is present in ``kwargs``.
    Throw ValueError if not.

    ``field_list`` can be a list/tuple of ``strings`` where each string is
    a field name or it can be a list/tuple of ``dicts`` where each item
    has the attributes ``name`` (**required**) and ``placeholder`` (**optional**).

    ``kwargs`` can be a dict or namedtuple. If a namedtuple, it calls its
    ``kwargs._as_dict()`` method and raises a ``ValueError`` if it does not succeed.

    * If the value of the item in ``kwargs`` is equal to its ``placeholder``
      defined in ``field_list``, a ``ValueError`` is raised.

    * If an item in ``kwargs`` is a *Resilient Select Function Input*, its
      value will be a ``dict`` that has a ``name`` attribute. This returns
      the value of ``name``.

    * If an item in ``kwargs`` is a *Resilient Multi-Select Function Input*, its
      value will be a list of ``dicts`` that have the ``name`` attribute. This
      returns a list of the ``name`` values for that item.

    :param field_list: the mandatory fields. *Can be an empty list if no mandatory fields.*
    :type field_list: list|tuple
    :param kwargs: dict or a namedtuple of all the fields to search.
    :type kwargs: dict|namedtuple
    :raises ValueError: if a field is missing
    :return: a Dictionary of all fields with Select/Multi-Select fields handled.
    :rtype: dict
    """

    mandatory_fields = field_list
    provided_fields = kwargs
    return_fields = {}
    mandatory_err_msg = "'{0}' is mandatory and is not set. You must set this value to run this function"

    # This is needed to handle something like: validate_fields(('incident_id'), kwargs)
    # In this case field_list will be a string and not a tuple
    if isinstance(mandatory_fields, string_types):
        mandatory_fields = [mandatory_fields]

    if not isinstance(mandatory_fields, list) and not isinstance(mandatory_fields, tuple):
        raise ValueError("'field_list' must be of type list/tuple, not {0}".format(type(mandatory_fields)))

    if not isinstance(provided_fields, dict):
        try:
            provided_fields = provided_fields._asdict()
        except AttributeError:
            raise ValueError("'kwargs' must be of type dict or namedtuple, not {0}".format(type(provided_fields)))

    # Validate that mandatory fields exist + are not equal to their placeholder values
    for field in mandatory_fields:

        placeholder_value = None

        if isinstance(field, dict):
            placeholder_value = field.get("placeholder")
            field = field.get("name")

        # If the field value is a defined empty str, raise an error
        if isinstance(provided_fields.get(field), string_types):
            if not provided_fields.get(field):
                raise ValueError(mandatory_err_msg.format(field))

        if provided_fields.get(field) is None:
            raise ValueError(mandatory_err_msg.format(field))

        if placeholder_value and provided_fields.get(field) == placeholder_value:
            raise ValueError(
                "'{0}' is mandatory and still has its placeholder value of '{1}'. You must set this value correctly to run this function".format(
                    field, placeholder_value))

    # Loop provided fields and get their value
    for field_name, field_value in provided_fields.items():

        # Handle if Select Function Input type
        if isinstance(field_value, dict) and field_value.get("name"):
            field_value = field_value.get("name")

        # Handle if 'Text with value string Input' type
        elif isinstance(field_value, dict) and field_value.get("content"):
            field_value = field_value.get("content")

        # Handle if Multi-Select Function Input type
        # There is a chance the list has already been "normalized", so just append as is
        elif isinstance(field_value, list):
            field_value = [f.get("name") if isinstance(f, dict) else f for f in field_value]

        return_fields[field_name] = field_value

    return return_fields


def get_file_attachment(res_client, incident_id, artifact_id=None, task_id=None, attachment_id=None):
    """
    Call the SOAR REST API to get the attachment or artifact data for
    an Incident or a Task

    * If ``incident_id`` and ``artifact_id`` are defined it will get that Artifact
    * If ``incident_id`` and ``attachment_id`` are defined it will get that Incident Attachment
    * If ``incident_id``, ``task_id`` and ``attachment_id`` are defined it will get that Task Attachment

    .. note::
        The ``artifact_id`` must reference an Artifact that is a downloadable type or it will
        raise a ``resilient.SimpleHTTPException``

    **Example:**

    .. code-block:: python

        artifact_data = get_file_attachment(self.rest_client(), incident_id=2001, artifact_id=1)

        with open("malware.eml", "wb") as f:
            f.write(artifact_data)

    :param res_client: required for communication back to SOAR
    :type res_client: resilient_circuits.ResilientComponent.rest_client()
    :param incident_id: id of the Incident
    :type incident_id: int|str
    :param artifact_id: id of the Incident's Artifact to download
    :type artifact_id: int|str
    :param task_id: id of the Task to download it's Attachment from
    :type task_id: int|str
    :param attachment_id: id of the Incident's Attachment to download
    :type attachment_id: int|str
    :return: byte string of attachment
    :rtype: str
    """

    if incident_id and artifact_id:
        data_uri = "/incidents/{}/artifacts/{}/contents".format(incident_id, artifact_id)
    elif attachment_id:
        if task_id:
            data_uri = "/tasks/{}/attachments/{}/contents".format(task_id, attachment_id)
        elif incident_id:
            data_uri = "/incidents/{}/attachments/{}/contents".format(incident_id, attachment_id)
        else:
            raise ValueError("task_id or incident_id must be specified with attachment")
    else:
        raise ValueError("artifact or attachment or incident id must be specified")

    # Get the data
    return res_client.get_content(data_uri)


def get_file_attachment_metadata(res_client, incident_id, artifact_id=None, task_id=None, attachment_id=None):
    """
    Call the SOAR REST API to get the attachment or artifact attachment metadata

    :param res_client: required for communication back to SOAR
    :type res_client: resilient_circuits.ResilientComponent.rest_client()
    :param incident_id: id of the Incident
    :type incident_id: int|str
    :param artifact_id: id of the Incident's Artifact
    :type artifact_id: int|str
    :param task_id: id of the Task to get it's Attachment metadata
    :type task_id: int|str
    :param attachment_id: id of the Incident's Attachment
    :type attachment_id: int|str
    :return: File metadata returned from endpoint
    :rtype: dict
    """

    if incident_id and artifact_id:
        metadata_url = "/incidents/{}/artifacts/{}".format(incident_id, artifact_id)
        return res_client.get(metadata_url)["attachment"]

    if attachment_id:
        if task_id:
            metadata_url = "/tasks/{}/attachments/{}".format(task_id, attachment_id)
        elif incident_id:
            metadata_url = "/incidents/{}/attachments/{}".format(incident_id, attachment_id)
        else:
            raise ValueError("If attachment_id is defined, you must specify task_id OR incident_id")

        return res_client.get(metadata_url)

    raise ValueError("artifact_id AND incident_id, OR attachment_id AND (task_id OR incident_id) must be specified")


def get_file_attachment_name(res_client, incident_id=None, artifact_id=None, task_id=None, attachment_id=None):
    """
    Call the SOAR REST API to get the attachment or artifact attachment name

    :param res_client: required for communication back to SOAR
    :type res_client: resilient_circuits.ResilientComponent.rest_client()
    :param incident_id: id of the Incident
    :type incident_id: int|str
    :param task_id: id of the Task
    :type task_id: int|str
    :param attachment_id: id of the Incident's Attachment to get the name for
    :type attachment_id: int|str
    :return: name of the Attachment or Artifact
    :rtype: str
    """

    name = ""
    if incident_id and artifact_id:
        name_url = "/incidents/{}/artifacts/{}".format(incident_id, artifact_id)
        name = res_client.get(name_url)["attachment"]["name"]
    elif attachment_id:
        if task_id:
            name_url = "/tasks/{}/attachments/{}".format(task_id, attachment_id)
            name = res_client.get(name_url)["name"]
        elif incident_id:
            name_url = "/incidents/{}/attachments/{}".format(incident_id, attachment_id)
            name = res_client.get(name_url)["name"]
        else:
            raise ValueError("task_id or incident_id must be specified with attachment")
    else:
        raise ValueError("artifact or attachment or incident id must be specified")

    # Return name string
    return name


def write_file_attachment(res_client, file_name, datastream, incident_id, task_id=None, content_type=None):
    """
    Add a file attachment to SOAR using the REST API
    to an Incident or a Task

    **Example:**

    .. code-block:: python

        with open("malware.eml", "rb") as data_stream:
            res = write_file_attachment(self.rest_client(), "malware.eml", data_stream, 2001)

    :param res_client: required for communication back to SOAR
    :type res_client: :class:`ResilientComponent.rest_client() <resilient_circuits.actions_component.ResilientComponent.rest_client()>`
    :param file_name: name of the attachment to create
    :type file_name: str
    :param dataStream: stream of bytes used to create the attachment
    :type dataStream: stream of bytes
    :param incident_id: id of the Incident
    :type incident_id: int|str
    :param task_id: (optional) id of the Task
    :type task_id: int|str
    :param content_type: (optional) MIME type of attachment. Default is ``"application/octet-stream"``
    :param content_type: str
    :return: metadata of new attachment created
    :rtype: dict
    """

    content_type = content_type \
                   or mimetypes.guess_type(file_name or "")[0] \
                   or "application/octet-stream"

    """
    Writing to temp path so that the REST API client can use this file path
    to read and POST the attachment
    """

    # Create a new attachment by calling SOAR REST API

    if task_id:
        attachment_uri = "/tasks/{}/attachments".format(task_id)
    else:
        attachment_uri = "/incidents/{}/attachments".format(incident_id)

    new_attachment = res_client.post_attachment(attachment_uri,
                                                None,
                                                filename=file_name,
                                                mimetype=content_type,
                                                bytes_handle=datastream)

    if isinstance(new_attachment, list):
        new_attachment = new_attachment[0]

    return new_attachment


def readable_datetime(timestamp, milliseconds=True, rtn_format='%Y-%m-%dT%H:%M:%SZ'):
    """
    Convert an epoch timestamp to a string using a format

    :param timestamp: ts of object sent from SOAR Server i.e. ``incident.create_date``
    :type timestamp: int
    :param milliseconds: Set to ``True`` if ts in milliseconds
    :type milliseconds: bool
    :param rtn_format: Format of resultant string. See https://docs.python.org/3.6/library/datetime.html#strftime-and-strptime-behavior for options
    :type rtn_format: str
    :return: string representation of timestamp
    :rtype: str
    """
    if milliseconds:
        ts = int(timestamp / 1000)
    else:
        ts = timestamp

    return datetime.datetime.utcfromtimestamp(ts).strftime(rtn_format)


def str_to_bool(value):
    """
    Convert value to either a ``True`` or ``False`` boolean

    Returns ``False`` if ``value`` is anything
    other than: ``'1', 'true', 'yes' or 'on'``

    :param value: the value to convert
    :type value: str
    :return: ``True`` or ``False``
    :rtype: bool
    """
    value = str(value).lower()
    return value in ('1', 'true', 'yes', 'on')


def write_to_tmp_file(data, tmp_file_name=None, path_tmp_dir=None):
    """
    Writes data to a file in a safely created temp directory.

    * If no ``tmp_file_name`` is provided, a temp name will be given
    * If no ``path_tmp_dir`` is provided a temp directory is created with the prefix **resilient-lib-tmp-**

    .. note::

        When used, ensure you safely remove the created temp directory
        in the ``finally`` block of the ``FunctionComponent`` code.

    **Example:**

    .. code-block:: python

        import os
        import shutil

        try:
            path_tmp_file, path_tmp_dir = write_to_tmp_file(attachment_contents, tmp_file_name=attachment_metadata.get("name"))

        except Exception:
            yield FunctionError()

        finally:
            if path_tmp_dir and os.path.isdir(path_tmp_dir):
                shutil.rmtree(path_tmp_dir)

    :param data: bytes to be written to the file
    :type data: bytes
    :param tmp_file_name: name to be given to the file
    :type tmp_file_name: str
    :param path_tmp_dir: path to an existing directory to use as the temp dir
    :type path_tmp_dir: str
    :return: a tuple (path_tmp_file, path_tmp_dir)
    :rtype: tuple
    """

    # If no tmp_file_name provided use next tempfile candidate name
    if not tmp_file_name:
        tmp_file_name = next(tempfile._get_candidate_names())

    # If no path_tmp_dir provided, create one
    if not path_tmp_dir:
        path_tmp_dir = tempfile.mkdtemp(prefix="resilient-lib-tmp-")

    elif not os.path.isdir(path_tmp_dir):
        raise IOError("Path does not exist: {0}".format(path_tmp_dir))

    # Generate path to tmp file
    path_tmp_file = os.path.join(path_tmp_dir, tmp_file_name)

    # Write the file
    with io.open(path_tmp_file, mode="wb") as temp_file:
        temp_file.write(data)

    return (path_tmp_file, path_tmp_dir)


def close_incident(res_client, incident_id, kwargs, handle_names=False):
    """
    Close an incident in SOAR.

    * Any **required on close (roc)** fields that
      are needed, pass them as a ``field_name:field_value`` dict in ``kwargs``

    * If any **roc** select field needs to be identified as its name,
      set ``handle_names`` to ``True``

    **Example:**

    .. code-block:: python

        res = close_incident(
            res_client=self.rest_client(),
            incident_id=fn_inputs.incident_id,
            kwargs={"resolution_id": "Duplicate", "resolution_summary": "This ticket is a duplicate"},
            handle_names=True
        )

    :param res_client: required for communication back to SOAR
    :type res_client: resilient_circuits.ResilientComponent.rest_client()
    :param incident_id: id of the incident
    :type incident_id: int|str
    :param kwargs: required fields needed to close an incident in a ``field_name:field_value`` format
    :type kwargs: dict
    :param handle_names: if ``True``, any select field types in ``kwargs`` will take ``str`` instead of ``int`` as their value
    :type handle_names: bool
    :return: Response from the server indicating if the incident was closed or not
    :rtype: requests.Response
    """

    if not incident_id:
        raise ValueError("'incident_id' must be specified")

    # API call to the TypeRest for fields "required": "close" if not in kwargs throw an error
    required_fields = _get_required_fields(res_client)

    missing_fields = [field for field in required_fields if field not in kwargs]
    if missing_fields:
        raise ValueError("Missing mandatory field(s) to close an incident: {0}".format(missing_fields))

    # check for known mandatory field "plan_status" if not in kwargs add it
    mandatory_fields = kwargs.copy()
    if "plan_status" not in mandatory_fields:
        mandatory_fields["plan_status"] = "C"

    # API call to the SOAR REST API to patch the incident data (close incident)
    response = _patch_to_close_incident(res_client, incident_id, mandatory_fields, handle_names)

    return response


def _get_required_fields(res_client):
    """
    :param res_client: required for communication back to SOAR
    :return: list
    """
    fields = _get_incident_fields(res_client)
    fields_required = [field for field in fields if fields[field].get("required") == "close"]

    return fields_required


@cached(cache=TTLCache(maxsize=10, ttl=600))
def _get_incident_fields(res_client):
    """
    call the SOAR REST API to get list of fields required to close an incident
    this call is cached for multiple calls
    :param res_client: required for communication back to SOAR
    :return: json
    """
    uri = "/types/incident"
    response = res_client.get(uri)
    incident_fields = response.get("fields")

    return incident_fields


def _patch_to_close_incident(res_client, incident_id, close_fields, handle_names=False):
    """
    call the SOAR REST API to patch incident
    :param res_client: required for communication back to SOAR
    :param incident_id: required
    :param close_fields: required
    :return: response object
    """
    uri = "/incidents/{}".format(incident_id)

    if handle_names:
        uri = "{0}?handle_format=names".format(uri)

    previous_object = res_client.get(uri)
    patch = resilient.Patch(previous_object)

    for field in close_fields:
        patch.add_value(field, close_fields[field])

    response = res_client.patch(uri, patch)

    return response



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2020. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use

class IntegrationError(Exception):
    """
    Class used to signal Integration Errors. It doesn't add any specific information other than
    identifying the type of error

    .. code-block:: python

        from resilient_lib import IntegrationError

        raise IntegrationError("Example raising custom error")

    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2020. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use

from datetime import datetime
import platform
import pkg_resources  # part of setuptools

METRICS_VERSION = "1.0"

class FunctionMetrics:
    """
    Use this function to track metrics on the function's operation. It tracks information on the package,
    it's environment and the execution time
    """

    def finish(self):
        """ build out the final metrics data structure """
        self.end_time = datetime.now()

        ttl_time = self.end_time - self.start_time

        try:
            # get information about the package we're running
            pkg = pkg_resources.get_distribution(self.func)
        except pkg_resources.DistributionNotFound:
            pkg = MissingPkg()

        return {
            "version": METRICS_VERSION,
            "package": pkg.project_name,
            "package_version": pkg.version,
            "host": platform.node(),
            "execution_time_ms": int(ttl_time.total_seconds() * 1000),
            "timestamp": self.end_time.strftime("%Y-%m-%d %H:%M:%S")
        }

    def __init__(self, func):
        self.start_time = datetime.now()
        self.func = func


class MissingPkg:
    """ mock package use when the package name provided isn't found """
    project_name = "unknown"
    version = "unknown"




import datetime
import os

import jinja2
import pytest
from resilient_lib import (global_jinja_env, make_payload_from_template,
                           render, render_json)
from resilient_lib.components.templates_common import (soar_datetimeformat,
                                                       soar_splitpart,
                                                       soar_substitute,
                                                       soar_trimlist)

TEST_DATA = data = {
    "string": "templates",
    "number": 42,
    "numbers": [1, 2, 3],
    "value_list": [' a', ' b', ' c ', 'd'],
    "object": {"name": "v<a>lue"},
    "epochdate": 1020304050607,
    "utcdate": "2021-10-22T20:53:53.913Z",
    "iso8601": "2021-10-22T20:53:53+00:00",
    "domain": u"\u9ede\u770b",
    "url": "https://example.com:8080",
    "repeat_list1": [1, 2, 3, 2],
    "repeat_list2": [{"a": 1}, {"a": 2}, {"a": 3}, {"a": 2}],
    "norepeat_list3": [{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}],
    "json": {
        "something": 42,
        "something_else": [
            { "a": "aa"},
            { "b": "bb"}
        ]
    }
}

TEST_ATTRIBUTES = {
    "attributes": {
        "cn": ["Albert Einstein"],
        "createTimestamp": "2014-02-21 16:51:33+00:00",
        "creatorsName": "cn=admin,dc=example,dc=com",
        "entryCSN": ["20150720185447.990131Z#000000#000#000000"],
        "entryDN": "uid=einstein,dc=example,dc=com",
        "entryUUID": "29f6dc28-2f64-1033-898b-a53eb149a944",
        "hasSubordinates": False,
        "mail": ["einstein@ldap.forumsys.com"],
        "modifiersName": "cn=admin,dc=example,dc=com",
        "modifyTimestamp": "2015-07-20 18:54:47+00:00",
        "objectClass": ["inetOrgPerson", "organizationalPerson", "person", "top"],
        "sn": ["Einstein"],
        "structuralObjectClass": "inetOrgPerson",
        "subschemaSubentry": "cn=Subschema",
        "telephoneNumber": ["314-159-2653"],
        "uid": ["einstein"]
    },
    "dn": "uid=einstein,dc=example,dc=com"
}

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_OVERRIDE_TEMPLATE = os.path.join(CURRENT_DIR, "data/override_template.jinja")
TEST_DEFAULT_TEMPLATE = os.path.join(CURRENT_DIR, "data/default_template.jinja")

@pytest.mark.parametrize("template, expected_results", [
        ("thing", u"thing"),
        ("this is my {{number}}", u"this is my 42"),
        ("this is my {{string}}", u'this is my templates'),
        ("object.name is {{object.name}}", u"object.name is v<a>lue"),
        ("number is {{numbers[1]}}", u"number is 2"),
        ("numbers are {% for n in numbers%}{{n}}{% if not loop.last %},{% endif %}{% endfor %}", u"numbers are 1,2,3"),
    ])
def test_basic(template, expected_results):
    """Test some basic functionality

        Test the 'iso8601' filter
        assert test("Go back to {{epochdate|iso8601}}")
        u'Go back to 2002-05-02T01:47:30+00:00'
    """
    assert render(template, TEST_DATA) == expected_results

@pytest.mark.parametrize("template, expected_results", [
        ("{{string|upper}}", u'TEMPLATES'),
        ("{{string|replace('s', 'ss')}}", u'templatess'),
        ("{{numbers|last}}", '3')
    ])
def test_builtin_filters(template, expected_results):
    # make sure we still have access to built-in filters
    assert render(template, TEST_DATA) == expected_results

# todo maybe an error with default date_format
@pytest.mark.parametrize("template, expected_results", [
        ("{{utcdate|soar_datetimeformat(split_at='.')}}", "1634936033000"),
        ("{{iso8601|soar_datetimeformat(split_at='+', date_format='%Y-%m-%dT%H:%M:%S')}}", "1634936033000"),
        ("{{soar_datetimeformat('2002-05-02T01:47:30')}}", "1020304050000"),
        ("{{'1'|soar_substitute('{\"1\":\"Low\"}')}}", u"Low"),
        ("{{'2'|soar_substitute('{\"1\":\"Low\", \"DEFAULT\":\"High\"}')}}", u"High"),
        ("{{'abc - def'|soar_splitpart(1)}}", u"def"),
        ("{{'abc - def'|soar_splitpart(5)}}", u"abc - def"),
        ("{{'abc/def'|soar_splitpart(0, split_chars='/')}}", u"abc"),
        ("{{value_list|soar_trimlist}}", u"['a', 'b', 'c', 'd']"),
        ("{{object|js}}", '\n"name": "v<a>lue"\n'),
        ({"text":"<div>{{object.name|html}}</div>"}, u'{"text": "<div>v&lt;a&gt;lue</div>"}'),
        ("{{html('f<oo')}}", u'f&lt;oo'),
        ("{{url|url}}", 'https%3A//example.com%3A8080'),
        ("{{domain|idna}}", u'xn--c1yn36f'),
        ("{{domain|punycode}}", u'c1yn36f'),
        ("Go back to {{epochdate|iso8601}}", u'Go back to 2002-05-02T01:47:30+00:00'),
        ("{{repeat_list1|uniq}}", "[1, 2, 3]"),
        ('{{repeat_list2|uniq}}', "[{'a': 1}, {'a': 2}, {'a': 3}]"),
        ('{{norepeat_list3|uniq}}', "[{'a': 1}, {'a': 2}, {'a': 3}, {'a': 4}]"),
        ("this is my {{number|json}}", u"this is my 42"),
        ("this is my {{string|json}}", u'this is my "templates"'),
        ("{{string|base64}}", 'InRlbXBsYXRlcyI=\n')
    ])
def test_custom_filters(template, expected_results):
    assert render(template, TEST_DATA) == expected_results

def test_render():
    assert render("template {{value}}", {"value":"123"}) == u'template 123'

    assert render({"template": "{{value}}"}, {"value":"123"}) == u'{"template": "123"}'

    assert render('{"template": {{value|json}} }', {"value":'1"23'}) == u'{"template": "1\\"23" }'

    assert render('{"template": "{{value|js}}" }', {"value":'1"23'}) == u'{"template": "1\\"23" }'

    assert render('{"template": {{value|ldap}} }', {"value":'1*23'}) == u'{"template": 1\\2a23 }'

    assert render('shell "{{value|ps}}"', {"value":'$"foo"'}) == u'shell "`$`"foo`""'

    assert render('shell "{{value|sh}}"', {"value":'$"foo"'}) == u'shell "\\$\\"foo\\""'

    assert render('template={{value|timestamp}}', {"value":0}) == u'template=0'

    assert render('template={{value|timestamp}}', {}) == u'template=null'

    assert render('template={{value|timestamp}}', {"value":{"year":2015, "month":7, "day":15}}) == u'template=1436918400000'

    assert render('template={{value|timestamp}}', {"value":datetime.datetime(2015, 7, 15)}) == u'template=1436918400000'

    assert render('{"description": "DN={{dn}}, mail={{attributes.mail[0]}}"}', TEST_ATTRIBUTES) == \
        u'{"description": "DN=uid=einstein,dc=example,dc=com, mail=einstein@ldap.forumsys.com"}'

@pytest.mark.parametrize("template, data, expected_results", [
    (
        '{"result":"{{value}}"}',
        {"value": "the" + chr(10) + "new" + chr(10) + "thing"},
        {u'result': u'the new thing'}
    ),
    (
        '{"result":"{{value}}"}',
        {"value": "the" + chr(1) + "new" + chr(9) + "thing"},
        {u'result': u'the new thing'}
    )
])
def test_render_json(template, data, expected_results):
    assert render_json(template, data) == expected_results

@pytest.mark.parametrize("override_template, default_template, payload, return_json, expected_results", [
        (TEST_OVERRIDE_TEMPLATE, TEST_DEFAULT_TEMPLATE, TEST_ATTRIBUTES, True,
         {"entryDN": "uid=einstein,dc=example,dc=com", "createTimestamp": 1393001493000}),
        (TEST_OVERRIDE_TEMPLATE, None, TEST_ATTRIBUTES, True,
         {"entryDN": "uid=einstein,dc=example,dc=com", "createTimestamp": 1393001493000}),
        (None, TEST_DEFAULT_TEMPLATE, TEST_ATTRIBUTES, True,
         {u'creatorsName': u'cn=admin,dc=example,dc=com', u'hasSubordinates': False}),
        ("bad_path.jinja", TEST_DEFAULT_TEMPLATE, TEST_ATTRIBUTES, True,
         {u'creatorsName': u'cn=admin,dc=example,dc=com', u'hasSubordinates': False}),
        (None, TEST_DEFAULT_TEMPLATE, TEST_ATTRIBUTES, False,
         '{     "creatorsName": "cn=admin,dc=example,dc=com",     "hasSubordinates": false }')
])
def test_make_payload_from_template(override_template, default_template, payload, return_json, expected_results):
    assert make_payload_from_template(override_template,
                                      default_template,
                                      payload,
                                      return_json=return_json) == expected_results

@pytest.mark.parametrize("str_date, date_format, split_at, expected_results", [
    ("2022-21-02T07:36:17", "%Y-%d-%mT%H:%M:%S", None, 1645428977000),
    ("2022-21-02T07:36:17.124Z", "%Y-%d-%mT%H:%M:%S", ".", 1645428977000),
    ("2022-02-21T07:36:17.124Z", None, ".", 1645428977000),
    ("2022-02-21T07:36:17", None, None, 1645428977000)
])
def test_soar_datetimeformat(str_date, date_format, split_at, expected_results):
    if date_format:
        assert soar_datetimeformat(str_date, date_format=date_format, split_at=split_at) == expected_results
    else:
        assert soar_datetimeformat(str_date, split_at=split_at) == expected_results


@pytest.mark.parametrize("str_value, lookup, expected_results", [
    ("low", '{ "low": "_low", "medium": "_medium", "high": "_high", "DEFAULT": "_nf" }', "_low"),
    ("nf", '{ "low": "_low", "medium": "_medium", "high": "_high", "DEFAULT": "_nf" }', "_nf"),
    ("no default", '{ "low": "_low", "medium": "_medium", "high": "_high" }', "no default")
])
def test_soar_substitute(str_value, lookup, expected_results):
    assert soar_substitute(str_value, lookup) == expected_results

@pytest.mark.parametrize("str_value, index, split_chars, expected_results", [
    ("something - here", 0, None, "something"),
    ("something - here", 1, None, "here"),
    ("something - here", 1, None, "here")
])
def test_soar_splitpart(str_value, index, split_chars, expected_results):
    if split_chars:
        assert soar_splitpart(str_value, index, split_chars=split_chars) == expected_results
    else:
        assert soar_splitpart(str_value, index) == expected_results

@pytest.mark.parametrize("str_list, expected_results", [
    ("item1", ['item1']),
    ("item1,item2", ['item1', 'item2']),
    ("item1, item2", ['item1', 'item2']),
    (" item1, item2 ", ['item1', 'item2']),
    ("", ['']),
    (None, None)
])
def test_soar_trimlist(str_list, expected_results):
    if not isinstance(str_list, type(None)):
        assert soar_trimlist(str_list.split(",")) == expected_results
    else:
        assert soar_trimlist(str_list) == expected_results


def test_global_jinja_env():
    env = global_jinja_env()

    assert isinstance(env, jinja2.environment.Environment)
    assert env.filters.get("soar_trimlist")



import pytest

from resilient_lib.ui import Tab
from resilient_lib.ui.common import permission_to_edit

class TestEditPermissions(object):
    class TestTab(Tab):
        NAME = "TestTab"
        UUID = "42"
        SECTION = "fn_test"
        CONTAINS = []

    def test_function_permissions_lock_ui(self):
        opts = {
            "fn_test": {
                "ui_lock": "True"
            }
        }
        assert not permission_to_edit(TestEditPermissions.TestTab, opts)

    def test_integration_permissions_lock_ui(self):
        opts = {
            "integrations":{
                "ui_lock": "True"
            },
            "fn_test": {
            }
        }
        assert not permission_to_edit(TestEditPermissions.TestTab, opts)

    def test_global_permissions_lock_ui(self):
        opts = {
            "resilient": {
                "ui_lock": "True"
            }
        }
        assert not permission_to_edit(TestEditPermissions.TestTab, opts)

    def test_function_permissions_override_global_lock(self):
        opts = {
            "resilient":{
                "ui_lock": "True",
            },
            "fn_test": {
                "ui_lock": "False"
            }
        }
        assert permission_to_edit(TestEditPermissions.TestTab, opts)

    def test_function_permissions_override_global_unlock(self):
        opts = {
            "resilient":{
                "ui_lock": "False",
            },
            "fn_test": {
                "ui_lock": "True"
            }
        }
        assert not permission_to_edit(TestEditPermissions.TestTab, opts)

    def test_function_permissions_override_integrations_unlock(self):
        opts = {
            "resilient": {
                "ui_lock": "False"
            },
            "integrations":{
                "ui_lock": "False",
            },
            "fn_test": {
                "ui_lock": "True"
            }
        }
        assert not permission_to_edit(TestEditPermissions.TestTab, opts)

    def test_integrations_permissions_override_global_unlock(self):
        opts = {
            "resilient":{
                "ui_lock": "False",
            },
            "integrations": {
                "ui_lock": "True"
            }
        }
        assert not permission_to_edit(TestEditPermissions.TestTab, opts)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved

# resilient_common logging constants
WARN_BUILD_INCIDENT_ORG_ID = "Building an incident link without the ORG ID may produce a link that directs the user to a different incident resulting in unexpected results. It is recommended to provide the ORG ID." # pragma: no cover

MIN_SUPPORTED_PY_VERSION = (3, 6)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

"""Common paths used in tests"""

import os

SHARED_MOCK_DATA_DIR = os.path.dirname(os.path.realpath(__file__))
MOCK_CERTS_DIR = os.path.join(SHARED_MOCK_DATA_DIR, "mock_certs")

MOCK_CLIENT_CERT_FILE = os.path.join(MOCK_CERTS_DIR, "cert.pem")
MOCK_CLIENT_KEY_FILE = os.path.join(MOCK_CERTS_DIR, "key.open.pem")



# -*- coding: utf-8 -*-
# pragma pylint: disable=unused-argument, no-self-use
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved.

"""
Common methods and filters used to process 
`Jinja Templates <https://jinja.palletsprojects.com/en/3.1.x/>`_ 

**Usage:**

1. Create a Jinja template in the ``/util/templates`` directory of the App.

.. note::

    Normally a template is in JSON format:

    .. code-block::

        {
            "name": "{{ alert_name }}",
            "severity": "{{ alert_severity }}"
        }

2. (*Optional*) Add an option to your ``app.config`` file to specify an absolute
path to a custom Jinja template called: ``custom_template_path`` - if not provided the
value of ``DEFAULT_TEMPLATE_PATH`` is used.

3. Add the following code to your **Function**:

.. code-block:: python

    import pkg_resources
    from resilient_circuits import AppFunctionComponent, FunctionResult, app_function
    from resilient_lib import make_payload_from_template

    PACKAGE_NAME = "my_app"
    FN_NAME = "my_function_using_jinja"

    # Creating an absolute path to the template
    DEFAULT_TEMPLATE_PATH = pkg_resources.resource_filename(PACKAGE_NAME, "util/templates/<default_name>.jinja2")

    class FunctionComponent(AppFunctionComponent):

        def __init__(self, opts):
            super(FunctionComponent, self).__init__(opts, PACKAGE_NAME)

        @app_function(FN_NAME)
        def _app_function(self, fn_inputs):
            '''
            Function: A function showing how to use Jinja templates
            Inputs:
                -   fn_inputs.alert_severity
                -   fn_inputs.device_id
            '''

            custom_template_path = self.app_configs.custom_template_path if hasattr(self.app_configs, "custom_template_path") else ""

            template_rendered = make_payload_from_template(
                template_override=custom_template_path,
                default_template=DEFAULT_TEMPLATE_PATH,
                payload={
                    "alert_name": f"Malware found on device {fn_inputs.device_id}",
                    "alert_severity": fn_inputs.alert_severity
                },
                return_json=True
            )

            self.LOG.info(template_rendered)

            results = {"template_rendered": template_rendered}

            yield FunctionResult(results)

Output:

.. code-block:: python

    INFO [my_function_using_jinja] {'name': 'Malware found on device 303', 'severity': 'High'}

4. Update your ``README.md`` documentation file to include relevant information about the templates,
for example:

.. code-block::

    ## Templates for SOAR Cases
    It may necessary to modify the templates used to create SOAR cases based on a customer's required custom fields.
    Below are the default templates used which can be copied, modified and used with app.config's
    `custom_template_path` setting to override the default template.

    ### custom_template.jinja
    ```
    {
        "name": "{{ alert_name }}",
        "severity": "{{ alert_severity }}"
    }
    ```

======

"""

import calendar
import datetime
import json
import logging
import os
import pprint
import random
import re
import sys
import time

import pytz
from jinja2 import Environment, Undefined, select_autoescape
from jinja2.exceptions import TemplateError, TemplateSyntaxError
from resilient_lib import readable_datetime

if sys.version_info.major < 3:
    from cgi import escape as html_escape
else:
    # Python 3.2 adds html.escape() and deprecates cgi.escape().
    from html import escape as html_escape

if sys.version_info.major < 3:
    from base64 import encodestring as b64encode
else:
    # Python 3.x
    from base64 import encodebytes as b64encode

if sys.version_info.major < 3:
    from urllib import quote
else:
    # Python 3.x
    from urllib.parse import quote

LOG = logging.getLogger(__name__)

UNDEFINED_LABEL = "[undefined]"


def render(template, data):
    """
    Render data into a template, producing a string result. All the additional custom filters are available.

    :param template: Path to or a dict of the Jinja template
    :type template: str or dict
    :param data: JSON data to apply to the template
    :type data: dict
    :return: result from the rendering of the template. The template is usually a string, but can be a dict
    :rtype: str or dict

    **Examples:**

    .. code-block:: python

        >>> render("template {{value}}", {"value":"123"})
        u'template 123'

        >>> render({"template": "{{value}}"}, {"value":"123"})
        u'{"template": "123"}'

        You can escape values using the 'json' filter,
        or the 'url' or 'html' or 'ldap' filters.

        >>> render('{"template": {{value|json}} }', {"value":'1"23'})
        u'{"template": "1\\\\"23" }'

        >>> render('{"template": "{{value|js}}" }', {"value":'1"23'})
        u'{"template": "1\\\\"23" }'

        >>> render('{"template": {{value|ldap}} }', {"value":'1*23'})
        u'{"template": 1\\\\2a23 }'

        >>> render('shell "{{value|ps}}"', {"value":'$"foo"'})
        u'shell "`$`"foo`""'

        >>> render('shell "{{value|sh}}"', {"value":'$"foo"'})
        u'shell "\\\\$\\\\"foo\\\\""'

        >>> render('template={{value|timestamp}}', {"value":0})
        u'template=0'

        >>> render('template={{value|timestamp}}', {})
        u'template=null'

        >>> render('template={{value|timestamp}}', {"value":{"year":2015, "month":7, "day":15}})
        u'template=1436918400000'

        >>> render('template={{value|timestamp}}', {"value":datetime.datetime(2015, 7, 15)})
        u'template=1436918400000'
    """

    stringtemplate = template
    if isinstance(template, dict):
        stringtemplate = json.dumps(template, sort_keys=True)

    try:
        jtemplate = environment().from_string(stringtemplate)
    except TemplateSyntaxError as err:
        LOG.error("Render failed: %s, with template: %s", str(err), stringtemplate)
        raise

    try:
        stringvalue = jtemplate.render(data)
    except TemplateError:
        LOG.error("Render failed, with data: %s", data)
        raise
    return stringvalue


def render_json(template, data):
    """
    Render data into a template, producing a JSON result.
    Also clean up any "really bad" control characters to avoid failure.

    :param template: Path to or a dict of the Jinja template
    :type template: str or dict
    :param data: dict to apply to the template
    :type data: dict
    :return: result from the rendering of the template as a dictionary
    :rtype: dict

    **Examples:**

    .. code-block:: python

       >>> d = {"value": "the" + chr(10) + "new" + chr(10) + "thing"}
       >>> render_json('{"result":"{{value}}"}', d)
       {u'result': u'the new thing'}

       >>> d = {"value": "the" + chr(1) + "new" + chr(9) + "thing"}
       >>> render_json('{"result":"{{value}}"}', d)
       {u'result': u'the new thing'}
    """
    result = render(template, data)
    result = _remove_ctl_chars(result)
    return _convert_to_json(result)


def _remove_ctl_chars(result):
    # replace any control characters with spaces
    for n in range(1, 32):
        result = result.replace(chr(n), " ")
    return result


def _convert_to_json(result):
    try:
        return json.loads(result)
    except:
        raise ValueError(u"It is expected that the rendered template is a JSON Object\nInvalid JSON result: {0}".format(result))


def make_payload_from_template(template_override, default_template, payload, return_json=True):
    """
    Convert a payload into a new format based on a specified template.

    :param template_override: Path to the specified template (*usually
        taken from the app.config file. See the Usage example above*)
    :type template_override: str
    :param default_template: Path to the default template (*usually in
        the '/util/templates' directory. See the Usage example above*)
    :type default_template: str
    :param payload: ``dict`` of payload that is passed to Jinja template
    :type payload: dict
    :param return_json: False if template should be render as a ``str``
        and results returned as a ``str``
    :type return_json: bool
    :return: If the Jinja template is valid JSON and ``return_json`` is ``True`` the result is
        returned as a ``dict`` else it returns the rendered template as a ``str``
    :rtype: str|dict
    :raises ValueError: if ``return_json`` is ``True`` and the Jinja template is not
        valid JSON
    """
    template_data = _get_template(template_override, default_template)

    # Render the template.
    if return_json:
        rendered_payload = render_json(template_data, payload)
    else:
        rendered_payload = render(template_data, payload)
        rendered_payload = _remove_ctl_chars(rendered_payload)
    LOG.debug(rendered_payload)

    return rendered_payload

def _get_template(specified_template, default_template):
    """return the contents of a jinja template, either from the default location or from a customer specified
        custom path

    Args:
        specified_template ([str]): [customer specified template path]
        default_template ([str]): [default template path]

    Returns:
        [str]: [contents of template]
    """
    template_file_path = specified_template
    if template_file_path:
        if not (os.path.exists(template_file_path) and os.path.isfile(template_file_path)):
            LOG.error(u"Template file: %s doesn't exist, using default template",
                        template_file_path)
            template_file_path = None

    if not template_file_path:
        # using default template
        template_file_path = os.path.join(
                                os.path.dirname(os.path.realpath(__file__)),
                                default_template
                            )

    LOG.debug(u"template file used: %s", template_file_path)
    with open(template_file_path, "r") as definition:
        return definition.read()

# C U S T O M   J I N J A   F I L T E R S


def soar_datetimeformat(value, date_format="%Y-%m-%dT%H:%M:%S", split_at=None):
    """
    **soar_datetimeformat**

    Convert UTC dates to epoch format.

    :param value: The UTC date string
    :type value: str
    :param date_format: *(optional)* Conversion format. Defaults to ``"%Y-%m-%dT%H:%M:%S"``
    :type date_format: str
    :param split_at: *(optional)* Character to split the date field to scope the date field

        .. code-block::

            split_at='.' to remove milliseconds for "2021-10-22T20:53:53.913Z"
            split_at='+' to remove tz information "2021-10-22T20:53:53+00:00"

    :type split_at: str

    :return: Epoch value of datetime, in milliseconds
    :rtype: int
    """
    if not value:
        return value

    if split_at:
        utc_time = time.strptime(value[:value.rfind(split_at)], date_format)
    else:
        utc_time = time.strptime(value, date_format)
    return calendar.timegm(utc_time) * 1000


def soar_substitute(value, json_str):
    """
    **soar_substitute**

    Replace values based on a lookup dictionary.

    :param value: original value to lookup
    :type value: str
    :param json_str: JSON encoded string with key/value pairs of lookup values
    :type json_str: JSON encoded str
    :return: replacement value or original value if no replacement found
    :rtype: str | int
    """
    replace_dict = json.loads(json_str)
    if value in replace_dict:
        return replace_dict[value]

    # use a default value if specific match is missing
    if 'DEFAULT' in replace_dict:
        return replace_dict['DEFAULT']

    return value


def soar_splitpart(value, index, split_chars=' - '):
    """
    **soar_splitpart**

    Split a string and return the index.

    :param value: string to split
    :type value: str
    :param index: index of split to return
    :type index: int
    :param split_chars: *(optional)* split characters.  Defaults to ``' - '``
    :type split_chars: str
    :return: value of split. If ``index`` is out of bounds, the original ``value`` is returned
    :rtype: str
    """
    splits = value.split(split_chars)
    if len(splits) > index:
        return splits[index]

    return value


def soar_trimlist(org_list):
    """
    **soar_trimlist**

    Trim whitespace from elements in a list.

    :param org_list: list of elements to trim whitespace from
    :type org_list: list of strings
    :return: list with elements trimmed of whitespace
    :rtype: list
    """
    if not isinstance(org_list, list):
        return org_list
    return [element.strip() for element in org_list]


def js_filter(val):
    """
    **js**

    Produces JSONified string of the value,
    without surrounding quotes.

    :param val: The string to convert
    :type val: str
    :return: JSONified string of the value, without surrounding quotes
    :rtype: str
    """
    if val is None or isinstance(val, Undefined):
        return "null"
    js = json_filter(val)
    return js[1:-1]


def json_filter(val, indent=0):
    """
    **json**

    Produces JSONified string of the value.

    :param val: The string to convert
    :type val: str
    :return: JSONified string of the value
    :rtype: str
    """
    if val is None or isinstance(val, Undefined):
        return "null"
    return json.dumps(val, indent=indent, sort_keys=True)


def html_filter(val):
    """
    **html**

    Produces HTML-encoded string of the value.

    :param val: The string to encode
    :type val: str
    :return: Encoded string
    :rtype: str
    """
    if isinstance(val, Undefined):
        return UNDEFINED_LABEL
    return html_escape(val)


def url_filter(val):
    """
    **url**

    Produces URL-encoded string of the value.

    :param val: The string to encoded
    :type val: str
    :return: Encoded string
    :rtype: str
    """
    if isinstance(val, Undefined):
        return UNDEFINED_LABEL
    return quote(str(val))


def idna_filter(val):
    """
    **idna**

    Encodes the value per RFC 3490.

    :param val: The string to encode
    :type val: str
    :return: Encoded string
    :rtype: str
    """
    if isinstance(val, Undefined):
        return UNDEFINED_LABEL
    return val.encode("idna").decode("utf-8")


def punycode_filter(val):
    """
    **punycode**

    Encodes the value per RFC 3492.

    :param val: The string to encode
    :type val: str
    :return: Encoded string
    :rtype: str
    """
    if isinstance(val, Undefined):
        return UNDEFINED_LABEL
    return val.encode("punycode").decode("utf-8")


def ldap_filter(val):
    """
    **ldap**

    Produces LDAP-encoded string of the value.

    :param val: The string to encode
    :type val: str
    :return: Encoded string
    :rtype: str
    """
    if isinstance(val, Undefined):
        return UNDEFINED_LABEL
    escaped = []
    for char in str(val):
        if char < '0' or char > 'z' or char in "\\*()":
            char = "\\%02x" % ord(char)
        escaped.append(char)
    return ''.join(escaped)


def ps_filter(val):
    """
    **ps**

    Escapes characters in ``val`` for use in a PowerShell command line.

    :param val: The string to escaped
    :type val: str
    :return: Escaped string
    :rtype: str
    """
    if isinstance(val, Undefined):
        return UNDEFINED_LABEL
    escaped = []
    for char in str(val):
        if char in "`$#'\"":
            char = "`" + char
        elif char == '\0':
            char = "`0"
        elif char == '\a':
            char = "`a"
        elif char == '\b':
            char = "`b"
        elif char == '\f':
            char = "`f"
        elif char == '\n':
            char = "`n"
        elif char == '\r':
            char = "`r"
        elif char == '\t':
            char = "`t"
        elif char == '\v':
            char = "`v"
        escaped.append(char)
    return ''.join(escaped)


def sh_filter(val):
    """
    **sh**

    Escapes characters in ``val`` for use in a Unix shell command line.

    :param val: The string to escaped
    :type val: str
    :return: Escaped string
    :rtype: str
    """
    if isinstance(val, Undefined):
        return UNDEFINED_LABEL
    escaped = []
    for char in str(val):
        if char in "$#\"":
            char = "\\" + char
        elif ord(char) < 32 or ord(char) > 126:
            char = "\\%03o" % ord(char)
        escaped.append(char)
    return ''.join(escaped)


def pretty_filter(val, indent=2):
    """
    **pretty**

    Produces pretty-printed string of the value.

    :param val: The string to format
    :type val: str
    :param indent: Number of tabs to use when formatting
    :type indent: int
    :return: The formatted string
    :rtype: str
    """
    if isinstance(val, Undefined):
        return UNDEFINED_LABEL

    def nice_repr(obj, context, maxlevels, level):
        if sys.version_info.major < 3:
            typ = type(obj)
            if typ is unicode:
                obj = obj.encode("utf-8")
        return pprint._safe_repr(obj, context, maxlevels, level)

    printer = pprint.PrettyPrinter(indent=indent)
    printer.format = nice_repr
    return printer.pformat(val)


def iso8601(val):
    """
    **iso8601**

    Assuming ``val`` is an epoch milliseconds timestamp, produce ISO8601 datetime.

    :param val: An epoch milliseconds timestamp
    :type val: str|int
    :return: ISO8601 datetime
    :rtype: str
    """
    dt = datetime.datetime.utcfromtimestamp(int(int(val)/1000))
    return pytz.UTC.localize(dt).isoformat()


def timestamp(val):
    """
    **timestamp**

    Try convert non-timestamp values to a timestamp.

    :param val: Either ``"now"`` or a dict containing year / month / day etc.
    :type val: str | dict
    :return: An epoch milliseconds timestamp
    :rtype: int

    .. code-block::

        >>> timestamp({"year": 2018, "month": 8, "day": 1, "timezoneID": "CET"})
        1533078000000

        >>> timestamp(Undefined())
        'null'

        >>> timestamp("now") > 1530000000000
        True

        >>> timestamp("now") > 2000000000000 # 2033
        False
    """
    if isinstance(val, dict):
        y = val.get("year", 1970)
        m = val.get("month", 1)
        d = val.get("day", 1)
        h = val.get("hour", 0)
        n = val.get("minute", 0)
        s = val.get("second", 0)
        u = val.get("milliSecond", 0)
        z = pytz.timezone(val.get("timezoneID", "UTC"))
        dt = datetime.datetime(y, m, d, h, n, s, u, z)
        return int(calendar.timegm(dt.utctimetuple()) * 1000)
    if isinstance(val, Undefined):
        return "null"
    if isinstance(val, datetime.datetime):
        return int(calendar.timegm(val.utctimetuple()) * 1000)
    if val == "now":
        return int(calendar.timegm(datetime.datetime.now().utctimetuple()) * 1000)
    return val


def uniq(val, key=None):
    """
    **uniq**

    Produce the unique list.  If ``val`` is a dict, produce unique list of key values.

    :param val: The original list
    :type val: [str | int | obj]
    :param key: If ``val`` is a dict return a list with dicts with just that ``key``
    :type key: str
    :return: Original list of items with duplicates removed
    :rtype: list

    .. code-block::

       >>> sorted(uniq([1,2,3,2]))
       [1, 2, 3]

       >>> sorted(uniq([ {"a":1}, {"a":2}, {"a":3}, {"a":2}]))
       [{'a': 1}, {'a': 2}, {'a': 3}]

       >>> sorted(uniq([ {"a":1}, {"a":2}, {"a":3}, {"a":2}, Exception()], "a"))
       [{'a': 1}, {'a': 2}, {'a': 3}, Exception()]
    """
    if not isinstance(val, list):
        return val
    if key is None:
        try:
            return list(set(val))
        except TypeError:
            pass
    keys = []
    values = []
    for value in val:
        try:
            thiskey = value[key]
        except:
            thiskey = repr(value)
        if thiskey not in keys:
            keys.append(thiskey)
            values.append(value)
    return values


def sample_filter(val, count=None):
    """
    **sample**

    Return a random sample from a list.

    :param val: List of str | obj | int
    :type val: list
    :param count: Number of times to repeat items in ``val`` to
        increase its *random* probability
    :type count: int | ``None``
    :return: The random item
    :rtype: str | obj | int
    """
    if count is None:
        # Return a single value
        try:
            return random.sample(list(val), 1)[0]
        except ValueError:
            return None
    else:
        # Return a list
        try:
            return random.sample(list(val), count)
        except ValueError:
            return []


def camel_filter(val):
    """
    **camel**

    Convert text to CamelCase.

    :param val: The string to convert
    :type val: str
    :return: Converted string
    :rtype: str

    .. code-block::

        This value is in camel case: {{ a#bc_def | camel }}
        >>> 'ABcDef'
    """
    titlecase = val.title()
    return re.sub(r"[\W^_]", "", titlecase)


def base64_filter(val, indent=2):
    """
    **base64**

    Breaks text into fixed-width blocks. You can specify the
    ``indent``.

    :param val: The string to convert
    :type val: str
    :param indent: Number of tabs
    :type indent: int
    :return: Converted string
    :rtype: str
    """
    if isinstance(val, Undefined):
        return ""
    s = json.dumps(val).encode("utf-8")
    return b64encode(s).decode("utf-8")


JINJA_FILTERS = {
    "json": json_filter,
    "js": js_filter,
    "html": html_filter,
    "url": url_filter,
    "idna": idna_filter,
    "punycode": punycode_filter,
    "ldap": ldap_filter,
    "ps": ps_filter,
    "sh": sh_filter,
    "pretty": pretty_filter,
    "timestamp": timestamp,
    "iso8601": iso8601,
    "uniq": uniq,
    "sample": sample_filter,
    "camel": camel_filter,
    "base64": base64_filter,
    "soar_datetimeformat": soar_datetimeformat,
    "soar_display_datetimeformat": readable_datetime,
    "soar_substitute": soar_substitute,
    "soar_splitpart": soar_splitpart,
    "soar_trimlist": soar_trimlist
}

# Maintain one global Environment
_ENV = Environment(autoescape=select_autoescape(default_for_string=False))
_ENV.globals.update(JINJA_FILTERS)
_ENV.filters.update(JINJA_FILTERS)


def global_jinja_env():
    """
    Return the Jinja environment with our resilient-lib custom filters.
    This environment can be expanded upon to add additional custom filters.

    See `Jinja Custom Filters <https://jinja.palletsprojects.com/en/3.1.x/api/#custom-filters>`_ for more.

    Current custom filters available:

    .. parsed-literal::

        |lib_jinja_filters|


    :return: The Jinja environment
    :rtype: `jinja2.Environment <https://jinja.palletsprojects.com/en/3.1.x/api/#jinja2.Environment>`_

    **Example:**

    .. code-block:: python

        from resilient-lib import global_jinja_env

        addl_custom_filters = {
            "filter_name": method_name
        }
        env = global_jinja_env()
        env.globals.update(addl_custom_filters)
        env.filters.update(addl_custom_filters)
    """
    return _ENV


environment = global_jinja_env



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2020. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use
from .common import create_tab
from .tab import Tab
from .elements import Datatable, Field, View



import pytest

from resilient_lib.ui import Tab, Datatable, Field

class TestTabMetaclass(object):
    """
    Confirm that a class can't be defined without all of the required
    class variables being selected.
    """

    def test_name_is_required(self):
        with pytest.raises(AttributeError, match="NAME"):
            class TestTab(Tab):
                UUID = "test"
                SECTION = "test"
                CONTAINS = []

    def test_uuid_is_required(self):
        with pytest.raises(AttributeError, match="UUID"):
            class TestTab(Tab):
                NAME = "test"
                SECTION = "test"
                CONTAINS = []

    def test_section_is_required(self):
        with pytest.raises(AttributeError, match="SECTION"):
            class TestTab(Tab):
                NAME = "test"
                UUID = "test"
                CONTAINS = []

    def test_contains_is_required(self):
        with pytest.raises(AttributeError, match="CONTAINS"):
            class TestTab(Tab):
                NAME = "test"
                UUID = "test"
                SECTION = "test"

class TestTab(object):
    class FakeTab(Tab):
        NAME = "Fake"
        UUID = "42"
        SECTION = "fake"

        CONTAINS = [
            Datatable("test_table1"),
            Datatable("test_table2"),
            Field("test_field1"),
            Field("test_field2")
        ]

    def test_exists_in(self):
        """
        This is a simplified version of "content" of a layout
        """
        assert TestTab.FakeTab.exists_in([
            {"predefined_uuid": "1"},
            {"predefined_uuid": "2"},
            {"predefined_uuid": "42"}
        ])

    def test_all_missing_fields(self):
        tabs = [
            {"predefined_uuid": "1"},
            {"predefined_uuid": "2"},
            {"predefined_uuid": "42",
             "fields": []
            }
        ]
        missing_fields = TestTab.FakeTab.get_missing_fields(tabs)
        assert len(missing_fields) == len(TestTab.FakeTab.CONTAINS)

    def test_all_missing_fields_with_other_existing(self):
        tabs = [
            {"predefined_uuid": "1"},
            {"predefined_uuid": "2"},
            {"predefined_uuid": "42",
             "fields": [
                 Datatable("unrelated").as_dto()
             ]
            }
        ]
        missing_fields = TestTab.FakeTab.get_missing_fields(tabs)
        assert len(missing_fields) == len(TestTab.FakeTab.CONTAINS)

    def test_one_missing_field(self):
        tabs = [
            {"predefined_uuid": "1"},
            {"predefined_uuid": "2"},
            {"predefined_uuid": "42",
             "fields": [
                 Datatable("test_table1").as_dto()
             ]
            }
        ]
        tab = TestTab.FakeTab.get_from_tabs(tabs)
        missing_fields = TestTab.FakeTab.get_missing_fields(tabs)
        assert len(missing_fields) == len(TestTab.FakeTab.CONTAINS)-1
        assert not Datatable("test_table1").exists_in(missing_fields)

    def test_one_missing_field_with_extra_fields(self):
        tabs = [
            {"predefined_uuid": "1"},
            {"predefined_uuid": "2"},
            {"predefined_uuid": "42",
             "fields": [
                 Datatable("test_table1").as_dto(),
                 Datatable("unrelated").as_dto()
             ]
            }
        ]
        tab = TestTab.FakeTab.get_from_tabs(tabs)
        missing_fields = TestTab.FakeTab.get_missing_fields(tabs)
        assert len(missing_fields) == len(TestTab.FakeTab.CONTAINS)-1
        assert not Datatable("test_table1").exists_in(missing_fields)



#!/usr/bin/env python

import pkg_resources

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    __version__ = None

pkg_resources.declare_namespace(__name__)

from resilient_lib.components.function_result import ResultPayload
from resilient_lib.components.html2markdown import MarkdownParser
from resilient_lib.components.requests_common import RequestsCommon, RequestsCommonWithoutSession
from resilient_lib.components.resilient_common import *
from resilient_lib.components.workflow_status import get_workflow_status
from resilient_lib.components.oauth2_client_credentials_session import OAuth2ClientCredentialsSession
from resilient_lib.components.integration_errors import IntegrationError
from resilient_lib.components.templates_common import *
from resilient_lib.components.poller_common import *



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.
import time
import logging
import requests

log = logging.getLogger(__name__)


class OAuth2ClientCredentialsSession(requests.Session):
    """
    Wrapper around `requests.Session <https://docs.python-requests.org/en/latest/api/#requests.Session>`_
    which receives authentication tokens through the ``client_credentials`` type of OAuth2 grant,
    and adds tokens to the requests made through the session.
    It also attempts to keep track of the token and refresh it as needed.

    * If proxies are defined, every request will use them.

    This session does not request authorization from the user first. The scope should be pre-authorized.

    **Example:**

    .. code-block:: python

        from resilient_lib import OAuth2ClientCredentialsSession, RequestsCommon

        api1 = OAuth2ClientCredentialsSession('https://example1.com/<tenant_id>/oauth/v2/', client_id='xxx', client_secret='xxx')
        api2 = OAuth2ClientCredentialsSession('https://example2.com/<tenant_id>/oauth/v2/', client_id='xxx', client_secret='xxx')

        api1.post('https://example1.com/v4/me/messages', data={}) # use as a regular requests session object

        api2.get('https://example2.com/v2/me/updates')

        # When writing an integration, use RequestsCommon to get the proxies defined in in your app.config file.
        rc = RequestsCommon(opts, function_opts)
        api3 = OAuth2ClientCredentialsSession('https://example3.com/{}/test', proxies=rc.get_proxies())

    :param url: Authorization URL, with tenant_id in it, if required
    :type url: str
    :param client_id: API key/User ID
    :type client_id: str
    :param client_secret: secret for API key
    :type client_secret: str
    :param scope: (optional) list of scopes
    :type scope: [str]
    :param proxies: (optional) Dictionary mapping protocol to the URL of the proxy.
            The mapping protocol must be in the format:

            .. code-block::

                {
                    "https_proxy": "https://localhost:8080,
                    "http_proxy": "http://localhost:8080
                }
    :type proxies: dict
    """

    AUTHORIZATION_ERROR_CODES = [401, 403]

    def __init__(self, url=None, client_id=None, client_secret=None, scope=None, proxies=None):
        """
        Get OAuth2 tokens and save them to be used in session requests.
        """
        # __init__ is run even if __new__ returned existing instance
        if getattr(self, "authorization_url", None) is not None:
            return

        super(OAuth2ClientCredentialsSession, self).__init__()
        log.debug("Creating OAuth2 Session.")
        if not url or not client_id or not client_secret:
            missing_fields = list()
            if not url:
                missing_fields.append("url")
            if not client_id:
                missing_fields.append("client_id")
            if not client_secret:
                missing_fields.append("client_secret")
            raise ValueError("Missing fields ({}) required for OAuth2 authentication."
                             .format(",".join(missing_fields)))

        self.authorization_url = url
        self.client_id = client_id
        self.client_secret = client_secret
        self.scope = scope

        self.oauth_token_data = None
        self.expires_in = None
        self.access_token = None
        self.token_type = None
        self.expiration_time = None
        self.proxies = proxies if proxies is not None else {}

        self.authenticate(url, client_id, client_secret, scope, proxies)

    def authenticate(self, url, client_id, client_secret, scope, proxies=None):
        """
        Authenticate with the endpoint
        """
        token_url = url

        log.debug("Requesting token from {0}".format(url))
        r = self.get_token(token_url, client_id, client_secret, scope, proxies)

        log.info("Response status code: {}".format(r.status_code))
        r.raise_for_status()

        self.oauth_token_data = r.json()
        log.debug("Received oauth token.")

        self.expires_in = self.oauth_token_data.get("expires_in", None)
        if self.expires_in is not None:
            self.expiration_time = time.time() + int(self.expires_in)
        else:
            self.expiration_time = None

        self.access_token = self.oauth_token_data["access_token"]
        self.token_type = self.oauth_token_data.get("token_type", None)

        return True

    def get_token(self, token_url, client_id, client_secret, scope=None, proxies=None):
        """
        **Override this method if the request needs specific information in the
        body of the request**

        For example, *Cloud Foundry* asking ``grant_type`` to be ``password`` and API key to be passed in ``password``.

        The default is:

        .. code-block:: python

            post_data = {
                'client_id': client_id,
                'client_secret': client_secret,
                'grant_type': 'client_credentials'
            }

            if scope:
                post_data['scope'] = scope

        :param token_url: Authorization URL, with tenant_id in it, if required
        :type token_url: str
        :param client_id: API key/User ID
        :type client_id: str
        :param client_secret: secret for API key
        :type client_secret: str
        :param scope: (optional) list of scopes
        :type scope: [str]
        :param proxies: (optional) Dictionary mapping protocol to the URL of the proxy.
        :type proxies: dict
        :return: the ``response`` from the ``token_url``
        :rtype: `requests.Response <https://docs.python-requests.org/en/latest/api/#requests.Response>`_
        """
        post_data = {
            'client_id': client_id,
            'client_secret': client_secret,
            'grant_type': 'client_credentials'
        }
        if scope:
            post_data['scope'] = scope
        return self.post(token_url, data=post_data, proxies=proxies)

    def update_token(self):
        """
        Institutes a request for a new access token.

        :raises ValueError: If it cannot update the token
        :return: True if could get a new token
        :rtype: bool
        """
        try:
            self.authenticate(self.authorization_url, self.client_id,
                              self.client_secret, self.scope, self.proxies)
        except ValueError:
            raise ValueError("Can't update the token, did the credentials for {0} change?"
                             .format(self.authorization_url))
        return True

    def request(self, method, url, *args, **kwargs):
        """Constructs a `requests.Request <https://docs.python-requests.org/en/latest/api/#requests.Request>`_,
        injects it with tokens and sends it in a new session to avoid having one session constantly open.

        This uses the ``requests.request()`` function to
        make a call. The inputs are mapped to this function.
        See `requests.request() <https://docs.python-requests.org/en/latest/api/#requests.request>`_
        for information on any parameters available

        :return: the ``response`` object
        :rtype: `requests.Response <https://docs.python-requests.org/en/latest/api/#requests.Response>`_
        """

        # Don't check expiration if attempting to refresh authorization - would cause infinite recursion loop
        if self.expiration_time is not None and url != self.authorization_url:
            if self.expiration_time < time.time():
                self.update_token()

        headers = kwargs.pop("headers") if "headers" in kwargs else {}
        proxies = kwargs.pop("proxies") if "proxies" in kwargs else self.proxies
        self.add_authorization_header(headers)

        resp = super(OAuth2ClientCredentialsSession, self).request(method, url, *args, headers=headers, proxies=proxies,
                                                                   **kwargs)

        # If the error anything other than Authorization issue, the problem is in user's request
        if resp.status_code in self.AUTHORIZATION_ERROR_CODES:
            self.update_token()
            resp = super(OAuth2ClientCredentialsSession, self).request(method, url, *args, headers=headers,
                                                                        proxies=proxies, **kwargs)

        return resp

    def add_authorization_header(self, headers):
        """
        Create headers needed for authentication/authorization, overriding the default ones if needed.

        :param headers: Dictionary containing the ``Authorization`` header
        :type headers: dict
        """
        headers["Authorization"] = "{} {}".format(self.token_type, self.access_token)



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.
import pytest
import requests
import time
from resilient_lib.components import oauth2_client_credentials_session
from resilient_lib.components.oauth2_client_credentials_session import OAuth2ClientCredentialsSession


def mock_token_request_factory(elapsed_time=None, expires_in=None, status_code=None, token=None):
    return lambda *args, **kwargs: MockTokenRequest(elapsed_time=elapsed_time, expires_in=expires_in,
                                                    status_code=status_code, token=token)


def raise_error(*args, **kwargs):
    raise ValueError()


def return_values_each_call(it):
    class StoreAndReturnValue(object):
        """
        Create a generator from iterator, and act like a function, returning
        next value every time class is called.
        """
        def __init__(self, it):
            self.gen = (x for x in it)
            self.current = 0

        def __call__(self, *args, **kwargs):
            self.current += 1
            return next(self.gen)

    return StoreAndReturnValue(it)


class MockTokenRequest(object):
    class ElapsedTime(object):
        def __init__(self, elapsed_time):
            self.elapsed_time = elapsed_time

        def total_seconds(self):
            return self.elapsed_time

    def __init__(self, elapsed_time=None, expires_in=None, status_code=None, token=None):
        self.status_code = status_code if status_code is not None else 200
        self.return_json = {
            "access_token": "token" if token is None else token
        }

        if elapsed_time is None:
            elapsed_time = 3
        self.elapsed = self.ElapsedTime(elapsed_time)

        if expires_in is not None:
            self.return_json["expires_in"] = expires_in

    def json(self):
        return self.return_json

    def raise_for_status(self):
        if self.status_code > 299:
            raise ValueError()


class TestAuthenticator(object):
    @pytest.fixture(autouse=True)
    def clean_up_singletons(self):
        if OAuth2ClientCredentialsSession.__dict__.get("__it__", None) is not None:
            del OAuth2ClientCredentialsSession.__it__

    def test_multiple_sessions_multiple_tokens(self, monkeypatch):
        monkeypatch.setattr(requests.sessions.Session, 'post', mock_token_request_factory(token="api1"))
        api1 = OAuth2ClientCredentialsSession("url:example1", "s", "t")
        monkeypatch.setattr(requests.sessions.Session, 'post', mock_token_request_factory(token="api2"))
        api2 = OAuth2ClientCredentialsSession("url:example2", "s", "t")

        assert api1.access_token == "api1"
        assert api2.access_token == "api2"

    def test_creates_session_with_passing_authentication(self, monkeypatch):
        monkeypatch.setattr(requests.sessions.Session, 'post', mock_token_request_factory())
        auth = OAuth2ClientCredentialsSession("test", "test", "test")
        assert getattr(auth, 'access_token', None) is not None

    def test_fails_to_create_session_with_bad_authentication(self, monkeypatch):
        monkeypatch.setattr(requests.sessions.Session, 'post', mock_token_request_factory(status_code=403))
        with pytest.raises(ValueError):
            OAuth2ClientCredentialsSession("test", "test", "test")

    def test_authenticator_fails_without_required_fields(self, monkeypatch):
        monkeypatch.setattr(OAuth2ClientCredentialsSession, 'authenticate', lambda *_: True)
        with pytest.raises(ValueError):
            OAuth2ClientCredentialsSession(None, None, None)
        with pytest.raises(ValueError):
            OAuth2ClientCredentialsSession(None, "s", "t")
        with pytest.raises(ValueError):
            OAuth2ClientCredentialsSession("t", None, "t")
        with pytest.raises(ValueError):
            OAuth2ClientCredentialsSession("t", "s", None)

    def test_scope_gets_passed_when_given(self, monkeypatch):
        def confirm_scope_data(*args, **kwargs):
            data = kwargs.pop("data", None)  # instead of data in signature because of 2.7
            assert data is not None
            assert data['scope'] is not None
            return mock_token_request_factory()()

        monkeypatch.setattr(requests.sessions.Session, 'post', confirm_scope_data)
        OAuth2ClientCredentialsSession("test", "test", "test", scope=["scope"])

    def test_no_expiration_doesnt_fail(self, monkeypatch):
        monkeypatch.setattr(requests.sessions.Session, 'post', mock_token_request_factory())
        auth = OAuth2ClientCredentialsSession("test", "test", "test")
        assert auth.expiration_time is None
        assert getattr(auth, 'access_token', None) is not None

    def test_expiration_time_is_set(self, monkeypatch):
        monkeypatch.setattr(requests.sessions.Session, 'post', mock_token_request_factory(expires_in=50))
        auth = OAuth2ClientCredentialsSession("test", "test", "test")
        assert auth.expiration_time is not None

    def test_updating_token_after_expiration(self, monkeypatch):
        monkeypatch.setattr(requests.sessions.Session, 'post', mock_token_request_factory(expires_in=50))
        monkeypatch.setattr(OAuth2ClientCredentialsSession, 'update_token', raise_error)
        auth = OAuth2ClientCredentialsSession("test", "test", "test")
        auth.expiration_time = time.time() - 1

        with pytest.raises(ValueError):
            auth.get("https://www.ibm.com")

    def test_detect_updating_token_after_request_ends_after_expiration(self, monkeypatch):
        """
        Test that when a request is made while the token is not expired, but the request fails,
        if sending the request was long enough where token might have expired, that the token will get refreshed,
        and the request will be made again.

        monkeypatch 'post', because authentication uses POST request
        monkey patch module's import of time, because pytest also uses time, to create the conditions
        where token expired during the duration of request being made.
        """
        monkeypatch.setattr(requests.sessions.Session, 'post',
                            mock_token_request_factory(expires_in=10, elapsed_time=0))
        monkeypatch.setattr(oauth2_client_credentials_session.time, 'time',
                            return_values_each_call([time.time(), time.time(), time.time() + 15]))
        monkeypatch.setattr(requests.sessions.Session, 'request',
                            mock_token_request_factory(status_code=401, elapsed_time=20))
        monkeypatch.setattr(OAuth2ClientCredentialsSession, 'update_token', raise_error)

        auth = OAuth2ClientCredentialsSession("test", "test", "test")

        with pytest.raises(ValueError):
            auth.get("https://www.ibm.com")
